-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      CipherCore for Deoxys
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;
use work.AES_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 32;    --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES      : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
    --! Determine the round size
    constant BS                     : integer := G_DBLK_SIZE;
    constant ADDR_WIDTH             : integer := 3+128/BS;
    constant ONES                   : std_logic_vector(BS       -1 downto 0)
                                        := (others => '1');
    constant ZEROS                  : std_logic_vector(BS       -1 downto 0)
                                        := (others => '0');
    constant BLOCK_NO_WIDTH         : integer := G_MAX_LEN-7;

    constant MSB_AD                 : std_logic_vector(3 downto 0) := "0010";
    constant MSB_AD_LAST            : std_logic_vector(3 downto 0) := "0110";
    constant MSB_M                  : std_logic_vector(3 downto 0) := "0000";
    constant MSB_M_LAST_NONZERO     : std_logic_vector(3 downto 0) := "0100";
    constant MSB_CHKSUM_FULL        : std_logic_vector(3 downto 0) := "0001";
    constant MSB_CHKSUM_NON_FULL    : std_logic_vector(3 downto 0) := "0101";

    type     rcon_type  is array (0 to 15) of std_logic_vector(  7 downto 0);
    type     ram_type   is array (0 to 15) of std_logic_vector(127 downto 0);
    constant ROUNDS                 : integer                      := 14;
    constant INV_OFFSET             : integer                      := 3;
    constant rcon_rom               : rcon_type := (x"5E",x"BC",x"63",x"C6",
                                                x"97",x"35",x"6A",x"D4",
                                                x"B3",x"7D",x"FA",x"EF",
                                                x"C5",x"91",x"39",x"72");
    --! =======================================================================
    --! Datapath
    --! =======================================================================
    --! Registers
    signal checksum         : std_logic_vector(BS               -1 downto 0);
    signal auth             : std_logic_vector(BS               -1 downto 0);
    signal state            : std_logic_vector(BS               -1 downto 0);
    signal rdkey_init       : std_logic_vector(BS               -1 downto 0);
    signal bdi_vbytes_r     : std_logic_vector(BS/8             -1 downto 0);
    signal bdi_ploc_r       : std_logic_vector(BS/8             -1 downto 0);
    signal key_r            : std_logic_vector(128              -1 downto 0);
    signal npub_r           : std_logic_vector(BS/2             -1 downto 0);
    signal tweakey_r        : std_logic_vector(BS*2             -1 downto 0);

    --! Tweakey and RoundKey signals
    signal npub_sel         : std_logic_vector(64               -1 downto 0);
    signal block_no         : std_logic_vector(BLOCK_NO_WIDTH   -1 downto 0);
    signal block_no_m1      : std_logic_vector(BLOCK_NO_WIDTH   -1 downto 0);
    signal block_no_p1      : std_logic_vector(BLOCK_NO_WIDTH   -1 downto 0);
    signal block_no_sel     : std_logic_vector(BLOCK_NO_WIDTH   -1 downto 0);
    signal rdkey_ram1       : ram_type;
    signal rdkey_ram2       : ram_type;

    signal tweakey_init     : std_logic_vector(BS*2             -1 downto 0);
    signal new_tweakey      : std_logic_vector(BS*2             -1 downto 0);
    signal to_tweakey       : std_logic_vector(BS*2             -1 downto 0);

    --! Round key
    signal rcon             : std_logic_vector(8                -1 downto 0);
    signal rcon_ext         : std_logic_vector(32               -1 downto 0);
    signal new_rdkey        : std_logic_vector(BS               -1 downto 0);

    --! Expanded status val
    signal valid_bits       : std_logic_vector(BS               -1 downto 0);
    signal pad_loc          : std_logic_vector(BS               -1 downto 0);

    --! to and from core operation
    signal input_data       : std_logic_vector(BS               -1 downto 0);
    signal rdkey1           : std_logic_vector(BS               -1 downto 0);
    signal rdkey2           : std_logic_vector(BS               -1 downto 0);
    signal rdkey            : std_logic_vector(BS               -1 downto 0);
    signal from_round_enc   : std_logic_vector(BS               -1 downto 0);
    signal from_round_dec   : std_logic_vector(BS               -1 downto 0);
    signal state_map        : t_AES_state;
    signal rdkey_map        : t_AES_state;
    signal from_round_enc_map : t_AES_state;
    signal from_round_dec_map : t_AES_state;

    --! Output
    signal output           : std_logic_vector(BS               -1 downto 0);
    signal tagout           : std_logic_vector(BS               -1 downto 0);

    --! =======================================================================
    --! Controller
    --! =======================================================================
    signal rnd              : std_logic_vector(ADDR_WIDTH       -1 downto 0);
    signal rnd_r            : std_logic_vector(ADDR_WIDTH       -1 downto 0);
    signal inv_rnd          : std_logic_vector(ADDR_WIDTH       -1 downto 0);
    signal inv_rnd_m1       : std_logic_vector(ADDR_WIDTH       -1 downto 0);

    signal is_decrypt       : std_logic;
    signal is_wr_ram2       : std_logic;    --! Status
    signal is_nonfull_vld   : std_logic;    --! Status non-full ready
    signal is_msg           : std_logic;    --! Status message data
    signal is_full          : std_logic;    --! Status full block
    signal is_auth          : std_logic;    --! Status auth data
    signal is_ovalid        : std_logic;    --! Status output valid

    signal ld_auth          : std_logic;
    signal ld_chksum        : std_logic;
    signal ld_rnd           : std_logic;
    signal wr_first         : std_logic;
    signal en_auth          : std_logic;
    signal en_block         : std_logic;
    signal en_chksum        : std_logic;
    signal en_ram           : std_logic;
    signal en_rnd           : std_logic;
    signal en_tweakey       : std_logic;
    signal sel_tweakey      : std_logic;
    signal en_state         : std_logic;
    signal set_nonfull      : std_logic;
    signal sel_block        : std_logic_vector(2                -1 downto 0);
    signal sel_input        : std_logic_vector(2                -1 downto 0);
    signal sel_state        : std_logic_vector(2                -1 downto 0);
    signal sel_tag          : std_logic;
    signal sel_npub         : std_logic;
    signal stage_status     : std_logic_vector(4                -1 downto 0);

    signal set_auth         : std_logic;
    signal set_ovalid       : std_logic;
    signal set_decrypt      : std_logic;
    signal set_full         : std_logic;
    signal clr_full         : std_logic;
    signal clr_ovalid       : std_logic;
    signal op_decrypt       : std_logic; --! Perform decryption (for tweak)
    signal op_decrypt_r     : std_logic; --! Perform decryption (for core)
    signal en_key           : std_logic;
    signal en_npub          : std_logic;
    signal key_rdy          : std_logic;
    signal bdi_rdy          : std_logic;
    signal bdo_vld          : std_logic;

    --! Derived control signals
    signal addr1, addr2     : std_logic_vector(ADDR_WIDTH       -1 downto 0);

    --! State
    type state_type is (
        S_INIT,             S_LD_KEY,   S_LD_NPUB,
        S_LOAD,             S_PROCESS,  S_DONE);
    type phase_type is (
        PHASE_INIT, PHASE_AD, PHASE_MSG,
        PHASE_TAG_PREP, PHASE_TAG_INIT_PARTIAL, PHASE_TAG_INIT,
        PHASE_TAG, PHASE_TAG_DONE);
    signal cs, ns    : state_type;
    signal phase, next_phase  : phase_type;

    signal ld_auth_r        : std_logic;
    signal ld_chksum_r      : std_logic;
    signal en_auth_r        : std_logic;
    signal en_block_r       : std_logic;
    signal en_chksum_r      : std_logic;
    signal en_key_r         : std_logic;
    signal en_npub_r        : std_logic;
    signal en_ram_r         : std_logic;
    signal en_state_r       : std_logic;
    signal en_tweakey_r     : std_logic;
    signal wr_first_r       : std_logic;
    signal is_wr_ram2_r     : std_logic;
    signal is_full_r        : std_logic;
    signal save_info        : std_logic;
    signal sel_addr1_r      : std_logic;
    signal sel_addr2_r      : std_logic;
    signal sel_block_r      : std_logic_vector(2                -1 downto 0);
    signal sel_state_r      : std_logic_vector(2                -1 downto 0);
    signal sel_load_r       : std_logic;
    signal sel_input_r      : std_logic_vector(2                -1 downto 0);
    signal sel_npub_r       : std_logic;
    signal sel_tweakey_r    : std_logic;
    signal sel_tag_r        : std_logic;
    signal stage_status_r   : std_logic_vector(4                -1 downto 0);

    signal msg_adone        : std_logic;
    signal sel_load         : std_logic;
    signal sel_addr1        : std_logic;
    signal sel_addr2        : std_logic;

begin
    --! =======================================================================
    --! Datapath
    --! =======================================================================
    block_no_m1 <= std_logic_vector(
                    unsigned(block_no(BLOCK_NO_WIDTH-1 downto 0))
                    - to_unsigned(1,BLOCK_NO_WIDTH));
    block_no_p1 <= std_logic_vector(
                    unsigned(block_no(BLOCK_NO_WIDTH-1 downto 0))
                    + to_unsigned(1,BLOCK_NO_WIDTH));
    with sel_block_r select
    block_no_sel <= block_no_p1 when "00",
                    block_no when "01",
                    (others => '0') when others;

    prc:
    process(clk)
    begin
        if rising_edge(clk) then
            if (en_block_r = '1') then
                block_no <= block_no_sel;
            end if;

            --! Tweakey
            if (en_tweakey_r = '1') then
                tweakey_r    <= to_tweakey;
            end if;

            --! Round Keys
            if (wr_first_r = '1') then
                rdkey_init <= new_rdkey;
            end if;
            if (en_ram_r = '1') then
                if (is_wr_ram2_r = '0') then
                    rdkey_ram1(to_integer(unsigned(addr1))) <= new_rdkey;
                end if;
                if (is_wr_ram2_r = '1') then
                    rdkey_ram2(to_integer(unsigned(addr2))) <= new_rdkey;
                end if;
            end if;

            --! Check sum
            if (ld_chksum_r = '1') then
                checksum    <= (others => '0');
            elsif (en_chksum_r = '1') then
                if is_decrypt = '1' then
                    checksum <= checksum xor ((output and valid_bits) xor pad_loc);
                else
                    checksum <= checksum xor bdi;
                end if;
            end if;

            --! Auth
            if (ld_auth_r = '1') then
                auth        <= (others => '0');
            elsif (en_auth_r = '1') then
                auth        <= auth xor state;
            end if;

            --! State
            if (en_state_r = '1') then
                if (save_info = '1') then
                    bdi_vbytes_r    <= bdi_valid_bytes;
                    bdi_ploc_r      <= bdi_pad_loc;
                end if;
                if (sel_state_r = "00") then
                    state <= from_round_enc;
                elsif (sel_state_r = "01") then
                    state <= from_round_dec;
                else
                    state <= input_data xor rdkey_init;
                end if;
            end if;

            if (en_key_r = '1') then
                key_r       <= key_r(95 downto 0) & key;
            end if;

            if (en_npub_r = '1') then
                npub_r      <= bdi(BS-1 downto BS/2);
            end if;
        end if;
    end process;

    --! Tweakey
    npub_sel      <= (others => '0') when sel_npub_r = '1' else npub_r;
    tweakey_init  <= key_r & stage_status_r & npub_sel
        & ZEROS(128-stage_status'length-npub_sel'length
                -BLOCK_NO_WIDTH-1 downto 0)
        & block_no_sel;
    to_tweakey <= tweakey_init when sel_tweakey_r = '1' else new_tweakey;

    u_tweakey:
    entity work.TweakeyUpdate(structure)
    port map (ii => tweakey_r, oo => new_tweakey);

    --! Round key calculation
    rcon     <= x"2F" when sel_load_r = '1' else rcon_rom(to_integer(unsigned(rnd_r)));
    rcon_ext <= rcon & rcon & rcon & rcon;
    new_rdkey  <= (to_tweakey(BS+128-1 downto BS+96) xor to_tweakey(128-1 downto  96) xor x"01020408") &
                  (to_tweakey(BS+ 96-1 downto BS+64) xor to_tweakey( 96-1 downto  64) xor rcon_ext) &
                  (to_tweakey(BS+ 64-1 downto BS+32) xor to_tweakey( 64-1 downto  32)                ) &
                  (to_tweakey(BS+ 32-1 downto BS+ 0) xor to_tweakey( 32-1 downto   0)                );

    --! Round key (to Round operation)
    rdkey1 <= rdkey_ram1(to_integer(unsigned(addr1)));
    rdkey2 <= rdkey_ram2(to_integer(unsigned(addr2)));
    rdkey  <= rdkey1 when is_wr_ram2_r = '1' else rdkey2;

    gInfo:
    for i in BS/8-1 downto 0 generate
        valid_bits(7+(i*8)    downto i*8) <= (others => bdi_vbytes_r(i));
        pad_loc   (7+(i*8)              ) <= bdi_ploc_r(i);
        pad_loc   (6+(i*8)    downto i*8) <= (others => '0');
    end generate;

    --! State init
    with sel_input_r(1 downto 0) select
    input_data <= bdi               when "00",
                  (others => '0')   when "01",
                  checksum          when others;

    --! Round
    u_map1: entity work.AES_map(structure)
    port map ( ii => state, oo => state_map);
    u_map2: entity work.AES_map(structure)
    port map ( ii => rdkey, oo => rdkey_map);

    u_enc: entity work.AES_Round_Basic
    port map (
        din     => state_map,
        rkey    => rdkey_map,
        dout    => from_round_enc_map);
    u_dec: entity work.AES_InvRound_Basic
    port map (
        din     => state_map,
        rkey    => rdkey_map,
        dout    => from_round_dec_map);

    u_imap_enc: entity work.AES_invmap(structure)
    port map ( ii => from_round_enc_map, oo => from_round_enc);
    u_imap_dec: entity work.AES_invmap(structure)
    port map ( ii => from_round_dec_map, oo => from_round_dec);

    --! Output
    output         <= state when is_full_r = '1' else state xor bdi;
    tagout         <= state xor auth;
    bdo            <= tagout when sel_tag_r = '1' else output;
    msg_auth_valid <= '1' when tagout = bdi else '0';

    --! =======================================================================
    --! Controller
    --! =======================================================================
    process(clk)
    begin
        if rising_edge(clk) then
            ld_auth_r       <= ld_auth;
            ld_chksum_r     <= ld_chksum;
            en_auth_r       <= en_auth;
            en_block_r      <= en_block;
            en_chksum_r     <= en_chksum;
            en_key_r        <= en_key;
            en_npub_r       <= en_npub;
            en_ram_r        <= en_ram;
            en_state_r      <= en_state;
            en_tweakey_r    <= en_tweakey;

            wr_first_r      <= wr_first;
            is_wr_ram2_r    <= is_wr_ram2;
            is_full_r       <= is_full;
            save_info       <= bdi_rdy or sel_input(0);
            sel_addr1_r     <= sel_addr1;
            sel_addr2_r     <= sel_addr2;
            sel_block_r     <= sel_block;
            sel_state_r     <= sel_state;

            sel_load_r      <= sel_load;

            sel_input_r     <= sel_input;
            sel_npub_r      <= sel_npub;
            sel_tweakey_r   <= sel_tweakey;
            sel_tag_r       <= sel_tag;
            stage_status_r  <= stage_status;
            rnd_r           <= rnd;

            key_ready       <= key_rdy;
            bdi_ready       <= bdi_rdy;
            bdo_valid       <= bdo_vld;
            msg_auth_done   <= msg_adone;

            if (sel_addr1 = '1') then
                addr1 <= inv_rnd_m1;
            else
                addr1 <= rnd;
            end if;
            if (sel_addr2 = '1') then
                addr2 <= inv_rnd_m1;
            else
                addr2 <= rnd;
            end if;
        end if;
    end process;
    sel_load <= '1' when (cs = S_LOAD) else '0';
    sel_addr1 <= '1' when (is_wr_ram2 = '0' and (set_decrypt = '1' or (cs = S_PROCESS and op_decrypt = '1'))) else '0';
    sel_addr2 <= '1' when (is_wr_ram2 = '1' and (set_decrypt = '1' or (cs = S_PROCESS and op_decrypt = '1'))) else '0';

    inv_rnd <= rnd xor ONES(ADDR_WIDTH-1 downto 0);
    inv_rnd_m1 <= std_logic_vector(to_unsigned(ROUNDS-1,ADDR_WIDTH))
                  when sel_load = '1' else std_logic_vector(unsigned(inv_rnd) - INV_OFFSET);

    ctrlregs:
    process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                cs <= S_INIT;
            else
                cs <= ns;
                phase <= next_phase;
            end if;

            if (ld_rnd = '1') then
                rnd <= (others => '0');
            elsif (en_rnd = '1') then
                rnd <= std_logic_vector(unsigned(rnd) + 1);
            end if;

            if (cs = S_INIT) then
                is_wr_ram2 <= '0';
            elsif (cs = S_PROCESS and unsigned(rnd) = ROUNDS-1) then
                is_wr_ram2 <= not is_wr_ram2;
            end if;

            if (cs = S_INIT) then
                is_nonfull_vld <= '0';
            elsif (set_nonfull = '1') then
                is_nonfull_vld <= '1';
            elsif (cs = S_LOAD and ns = S_PROCESS) then
                is_nonfull_vld <= '0';
            end if;

            if (cs = S_LD_NPUB or phase = PHASE_TAG) then
                is_msg <= '0';
            elsif (bdi_rdy = '1') then
                if (phase = PHASE_MSG) then
                    is_msg <= bdi_size(G_LBS_BYTES);
                else
                    is_msg <= '0';
                end if;
            end if;

            if (cs = S_LD_NPUB) then
                is_decrypt <= decrypt;
            end if;

            if (clr_full = '1') then
                is_full <= '0';
            elsif (set_full = '1') then
                is_full <= '1';
            end if;
            if (set_auth = '1') then
                is_auth <= '1';
            elsif (en_auth = '1') then
                is_auth <= '0';
            end if;
            if (set_ovalid = '1') then
                is_ovalid <= '1';
            elsif (bdo_vld = '1' or clr_ovalid = '1') then
                is_ovalid <= '0';
            end if;

            if (cs = S_LD_NPUB) then
                op_decrypt <= '0';
            elsif (set_decrypt = '1') then
                op_decrypt <= '1';
            elsif (cs = S_PROCESS and unsigned(rnd) = ROUNDS-1) then
                op_decrypt <= '0';
                op_decrypt_r <= op_decrypt;
            end if;
        end if;
    end process;

    comb_tweakey:
    process(phase, bdi_eot, bdi_size, is_nonfull_vld, bdi_eoi, bdi_type)
    begin
        sel_npub     <= '0';
        stage_status <= (others => '0');
        case phase is
        when PHASE_INIT =>
            sel_npub     <= '1';
            stage_status <= MSB_AD;
        when PHASE_AD =>
            if (bdi_type(2 downto 1) = BDI_TYPE_ASS) then
                if (bdi_eot = '1') then
                    if (bdi_size(G_LBS_BYTES) = '1' or is_nonfull_vld = '1') then
                        if (bdi_eoi = '1') then
                            stage_status <= MSB_CHKSUM_FULL;
                        else
                            stage_status <= MSB_M;
                        end if;
                    else
                        sel_npub     <= '1';
                        stage_status <= MSB_AD_LAST;
                    end if;
                else
                    sel_npub     <= '1';
                    stage_status <= MSB_AD;
                end if;
            else
                if (bdi_size(G_LBS_BYTES) = '1') then
                    stage_status <= MSB_M;
                else
                    stage_status <= MSB_M_LAST_NONZERO;
                end if;
            end if;
        when PHASE_MSG =>
            if (bdi_eot = '1') then
                if (bdi_size(G_LBS_BYTES) = '1') then
                    stage_status <= MSB_CHKSUM_FULL;
                elsif (is_nonfull_vld = '1') then
                    stage_status <= MSB_CHKSUM_NON_FULL;
                else
                    stage_status <= MSB_M_LAST_NONZERO;
                end if;
            else
                stage_status <= MSB_M;
            end if;
        when PHASE_TAG_INIT_PARTIAL =>
            stage_status <= MSB_CHKSUM_NON_FULL;
        when PHASE_TAG_INIT =>
            stage_status <= MSB_CHKSUM_FULL;
        when OTHERS =>
        end case;
    end process;

    en_auth <= '1' when (cs = S_LOAD and is_auth = '1') else '0';

    comb:
    process(cs, phase,
        key_valid, key_update, bdo_ready,
        bdi_type, bdi_size, bdi_valid, bdi_eoi,
        op_decrypt_r,
        rnd, is_nonfull_vld, is_decrypt, is_msg, is_ovalid)
    begin
        ns              <= cs;
        --! External
        key_rdy         <= '0';
        bdi_rdy         <= '0';
        bdo_vld         <= '0';
        msg_adone       <= '0';
        --! Controls
        ld_auth         <= '0';
        ld_chksum       <= '0';
        ld_rnd          <= '0';
        wr_first        <= '0';
        en_block        <= '0';
        en_chksum       <= '0';
        en_ram          <= '0';
        en_rnd          <= '0';
        en_tweakey      <= '0';
        sel_tweakey     <= '0';
        en_state        <= '0';
        set_nonfull     <= '0';
        sel_block       <= "00";
        sel_input       <= "00";
        sel_state       <= "00";
        sel_tag         <= '0';
        en_key          <= '0';
        en_npub         <= '0';

        clr_ovalid      <= '0';
        clr_full        <= '0';
        set_ovalid      <= '0';
        set_full        <= '0';
        set_decrypt     <= '0';
        set_auth        <= '0';

        next_phase      <= phase;

        case cs is
        when S_INIT =>
            ld_rnd          <= '1';
            ld_chksum       <= '1';
            ld_auth         <= '1';
            next_phase      <= PHASE_INIT;
            sel_block       <= "10";
            en_block        <= '1';
            clr_full        <= '1';
            if (bdi_valid = '1') then
                if (key_update = '1') then
                    ns <= S_LD_KEY;
                else
                    ns <= S_LD_NPUB;
                end if;
            end if;

        when S_LD_KEY =>
            key_rdy <= '1';
            if (key_valid = '1') then
                en_key  <= '1';
                en_rnd  <= '1';
                if (unsigned(rnd) = 3) then
                    ns  <= S_LD_NPUB;
                end if;
            end if;

        when S_LD_NPUB =>
            ld_rnd     <= '1';
            bdi_rdy    <= '1';
            clr_ovalid  <= '1';
            if (bdi_valid = '1') then
                en_npub <= '1';
                ns      <= S_LOAD;
                if (bdi_eoi = '1') then
                    next_phase <= PHASE_TAG_INIT;
                end if;
            end if;

        when S_LOAD =>
            wr_first   <= '1';
            en_ram     <= '1';
            en_tweakey <= '1';
            sel_tweakey <= '1';
            sel_state(1) <= '1';

            case phase is
            when PHASE_INIT =>
                sel_block  <= "10";
                en_block   <= '1';
                ns         <= S_PROCESS;
                next_phase <= PHASE_AD;

            when PHASE_AD =>
                if (bdi_valid = '1') then
                    set_full <= '1';
                    if (bdi_type(2 downto 1) = BDI_TYPE_ASS) then
                        if (bdi_eot = '1') then
                            if bdi_size(G_LBS_BYTES) = '0' then
                                if (is_nonfull_vld = '1') then
                                    set_auth  <= '1';
                                    en_block  <= '1';
                                    sel_block <= "10";
                                    if (bdi_eoi = '1') then
                                        next_phase <= PHASE_TAG;
                                    else
                                        set_decrypt <= is_decrypt;
                                        next_phase <= PHASE_MSG;
                                    end if;
                                    bdi_rdy     <= '1';
                                    en_state    <= '1';
                                else
                                    sel_block   <= "01";
                                    set_nonfull <= '1';
                                end if;
                            else
                                set_auth  <= '1';
                                en_block  <= '1';
                                sel_block <= "10";
                                if (bdi_eoi = '1') then
                                    next_phase <= PHASE_TAG;
                                else
                                    set_decrypt <= is_decrypt;
                                    next_phase <= PHASE_MSG;
                                end if;
                                bdi_rdy  <= '1';
                                en_state <= '1';
                            end if;
                        else
                            set_auth <= '1';
                            en_block <= '1';
                            bdi_rdy  <= '1';
                            en_state <= '1';
                        end if;
                    else
                        --! Special transition when there's an empty AD
                        set_decrypt <= is_decrypt;
                        next_phase  <= PHASE_MSG;
                        en_block    <= '1';
                        sel_block   <= "10";
                    end if;
                    ns       <= S_PROCESS;
                end if;

            when PHASE_MSG =>
                if (bdi_valid = '1' and bdo_ready = '1') then
                    set_full <= '1';
                    if (is_ovalid = '1') then
                        bdo_vld <= '1';
                    end if;
                    if ((is_decrypt = '0' and bdi_eot = '0')
                            or (is_decrypt = '0' and bdi_eot = '1' and (bdi_size(G_LBS_BYTES) = '1' or is_nonfull_vld = '1'))
                            or (is_decrypt = '1' and  is_ovalid = '1')) then
                        en_chksum <= '1';
                    end if;
                    if (bdi_eot = '1') then
                        if bdi_size(G_LBS_BYTES) = '0' then
                            clr_full    <= '1';
                            if (is_nonfull_vld = '1') then
                                sel_input  <= "01";
                                next_phase <= PHASE_TAG_PREP;
                                en_state   <= '1';
                            else
                                sel_block   <= "01";
                                set_nonfull <= '1';
                            end if;
                            clr_ovalid <= '1';
                        else
                            en_block   <= '1';
                            set_ovalid <= '1';
                            next_phase <= PHASE_TAG_INIT;
                            bdi_rdy    <= '1';
                            en_state   <= '1';
                        end if;
                    else
                        set_decrypt <= is_decrypt;
                        en_block   <= '1';
                        set_ovalid <= '1';
                        bdi_rdy    <= '1';
                        en_state   <= '1';
                    end if;
                    ns <= S_PROCESS;
                end if;

            when PHASE_TAG_PREP =>
                if (bdo_ready = '1' and bdi_valid = '1') then
                    bdi_rdy   <= '1';
                    bdo_vld   <= '1';
                    en_chksum <= is_decrypt;
                    next_phase <= PHASE_TAG_INIT_PARTIAL;
                end if;

            when PHASE_TAG_INIT_PARTIAL | PHASE_TAG_INIT =>
                --! Special state for (empty AD and Msg) or (partial last message block)
                if (is_ovalid = '1') then
                    en_chksum <= is_decrypt;
                end if;
                sel_block   <= "01";    --! Dont get incremented block_no
                bdo_vld     <= is_ovalid;
                set_nonfull <= '1';
                ns          <= S_PROCESS;
                next_phase  <= PHASE_TAG;

            when PHASE_TAG =>       --! Starts final tag calculation
                sel_input <= "10";
                en_state  <= '1';
                ns <= S_PROCESS;
                next_phase <= PHASE_TAG_DONE;

            when PHASE_TAG_DONE =>  --! Output data and complete
                sel_tag <= '1';
                if (is_decrypt = '0') then
                    if (bdo_ready = '1') then
                        bdo_vld <= '1';
                        ns      <= S_DONE;
                    end if;
                else
                    if (bdi_valid = '1') then
                        bdi_rdy     <= '1';
                        msg_adone   <= '1';
                        ns          <= S_DONE;
                    end if;
                end if;
            end case;

        when S_PROCESS =>
            sel_state(1) <= '0';
            sel_state(0) <= op_decrypt_r;
            en_state <= not is_nonfull_vld;
            en_rnd   <= '1';
            en_tweakey  <= '1';
            if (unsigned(rnd) = ROUNDS-1) then
                ns      <= S_LOAD;
                ld_rnd  <= '1';
                if (is_decrypt = '1' and phase = PHASE_MSG
                        and bdi_size(G_LBS_BYTES) = '1') then
                -- if (op_decrypt = '1' and is_full = '1') then
                    wr_first <= '1';
                else
                    en_ram   <= '1';
                end if;
            else
                en_ram     <= '1';
            end if;

        when S_DONE =>
            ns <= S_INIT;
        end case;
    end process;

end structure;